r"""
Classes to provide insolation as input for other CLIMLAB processes.

Options include

- ``climlab.radiation.P2Insolation`` (idealized 2nd Legendre polynomial form)
- ``climlab.radiation.FixedInsolation`` (generic steady-state insolation)
- ``climlab.radiation.AnnualMeanInsolation`` (steady-state annual-mean insolation computed from orbital parameters and latitude)
- ``climlab.radiation.DailyInsolation`` (time-varying daily-mean insolation computed from orbital parameters, latitude and time of year)

All are subclasses of ``climlab.process.DiagnosticProcess``
and do add any tendencies to any state variables.

At least two diagnostics are provided:

- ``insolation``, the incoming solar radiation in :math:`\textrm{W}}{\textrm{m}^2}``
- ``coszen``, cosine of the solar zenith angle
"""
from __future__ import division
import numpy as np
from climlab.process.diagnostic import DiagnosticProcess
from climlab.domain.field import Field
from climlab.domain.field import to_latlon
from climlab.utils.legendre import P2
from climlab import constants as const
from climlab.solar.insolation import daily_insolation

# REVISE TO MAKE ALL OF THESE CALLABLE WITH NO ARGUMENTS.
# SET SOME SENSIBLE DEFAULTS FOR DOMAINS

#  the diagnostic self.insolation is set with correct dimensions at
#  creation time. After that, make sure to always modify it though
#  self.insolation[:] = ...
#  so that links to the insolation in other processes will work

#  REALLY NEED TO CHANGE THE WAY DIAGNOSTIC PROCESSES ARE CREATED
#  CAN'T RELY ON NAMES OF DOMAINS
#  should be easy to pass a state variable object or something with the right shape
#  and have the process gracefully set the correct dimensions


class _Insolation(DiagnosticProcess):
    """A private parent class for insolation processes.

    Calling compute() will update self.insolation with current values.

    **Initialization parameters** \n

    An instance of ``_Insolation`` is initialized with the following
    arguments *(for detailed information see Object attributes below)*:

    :param float S0:        solar constant                              \n
                            - unit: :math:`\\frac{\\textrm{W}}{\\textrm{m}^2}`   \n
                            - default value: ``1365.2``

    **Object attributes** \n

    Additional to the parent class :class:`~climlab.process.diagnostic.DiagnosticProcess`
    following object attributes are generated and updated during initialization:

    :ivar array insolation: the array is initialized with zeros of the size of
                            ``self.domains['sfc']`` or ``self.domains['default']``.
    :ivar array coszen:     cosine of the solar zenith angle
    :ivar float S0:         initialized with given argument ``S0``
    :ivar dict diagnostics: key ``'insolation'`` initialized with value:
                            :class:`~climlab.domain.field.Field` of zeros
                            in size of ``self.domains['sfc']`` or
                            ``self.domains['default']``
    :ivar Field insolation: the subprocess attribute ``self.insolation`` is
                            created with correct dimensions

    .. note::

        ``self.insolation`` should always be modified with
        ``self.insolation[:] = ...`` so that links to the insolation in other
        processes will work.

    """
    # parameter S0 is now stored using a python property
    # can be changed through self.S0 = newvalue
    # which will also update the parameter dictionary
    # CAUTION: changing self.param['S0'] will not work!
    def __init__(self, S0=const.S0, **kwargs):
        super(_Insolation, self).__init__(**kwargs)
        #  initialize diagnostics with correct shape
        self.add_diagnostic('insolation')
        self.add_diagnostic('coszen')
        try:
            domain = self.domains['sfc']
        except:
            domain = self.domains['default']
        self.insolation = Field(np.zeros(domain.shape), domain=domain)
        self.coszen = Field(np.zeros(domain.shape), domain=domain)
        self.declare_diagnostics(['insolation','coszen'])
        self.S0 = S0
        #  Now that we have a value for self.S0 we can compute the correct coszen
        self.coszen = self._coszen_from_insolation()
        self.declare_input(['S0'])

    @property
    def S0(self):
        """Property of solar constant S0.

        The parameter S0 is stored using a python property and can be changed through
        ``self.S0 = newvalue`` which will also update the parameter dictionary.

        .. warning::

            changing ``self.param['S0']`` will not work!

        :getter:    Returns the S0 parameter which is stored in attribute ``self._S0``.
        :setter:    * sets S0 which is addressed as ``self._S0`` to the new value
                    * updates the parameter dictionary ``self.param['S0']`` and
                    * calls method :func:`_compute_fixed`
        :type:      float

        """
        return self._S0
    @S0.setter
    def S0(self, value):
        self._S0 = value
        self.param['S0'] = value
        self._compute_fixed()

    def _coszen_from_insolation(self):
        return self.insolation / self.S0

    def _compute_fixed(self):
        '''Recompute any fixed quantities after a change in parameters'''
        pass

    def _get_current_insolation(self):
        pass

    def _compute(self):
        self._get_current_insolation()
        return {}

class FixedInsolation(_Insolation):
    """A class for fixed insolation at each point of latitude off the domain.

    The solar distribution for the whole domain is constant and specified by
    a parameter.

    **Initialization parameters** \n

    :param float S0:        solar constant                              \n
                            - unit: :math:`\\frac{\\textrm{W}}{\\textrm{m}^2}`   \n
                            - default value: ``const.S0/4 = 341.2``

    :Example:

        ::

            >>> import climlab
            >>> from climlab.radiation.insolation import FixedInsolation

            >>> model = climlab.EBM()
            >>> sfc = model.Ts.domain

            >>> fixed_ins = FixedInsolation(S0=340.0, domains=sfc)

            >>> print fixed_ins
            climlab Process of type <class 'climlab.radiation.insolation.FixedInsolation'>.
            State variables and domain shapes:
            The subprocess tree:
            top: <class 'climlab.radiation.insolation.FixedInsolation'>

    """
    def __init__(self, S0=const.S0/4, **kwargs):
        super(FixedInsolation, self).__init__(S0=S0, **kwargs)

    def _compute_fixed(self):
        #ins_adjustment = self.S0 - self.insolation
        #self.insolation += ins_adjustment
        self.insolation[:] = self.S0
        self.coszen[:] = self._coszen_from_insolation()


class P2Insolation(_Insolation):
    """A class for parabolic solar distribution over the domain's latitude
    on the basis of the second order Legendre Polynomial.

    Calculates the latitude dependent solar distribution as

    .. math::

        S(\\varphi) = \\frac{S_0}{4} \\left( 1 + s_2 P_2(x) \\right)

    where :math:`P_2(x) = \\frac{1}{2} (3x^2 - 1)` is the second order Legendre Polynomial
    and :math:`x=sin(\\varphi)`.

    **Initialization parameters** \n

    :param float S0:        solar constant                              \n
                            - unit: :math:`\\frac{\\textrm{W}}{\\textrm{m}^2}`   \n
                            - default value: ``1365.2``
    :param floar s2:        factor for second legendre polynominal term \n
                            - default value: ``-0.48``

    :Example:

        ::

            >>> import climlab
            >>> from climlab.radiation.insolation import P2Insolation

            >>> model = climlab.EBM()
            >>> sfc = model.Ts.domain

            >>> p2_ins = P2Insolation(S0=340.0, s2=-0.5, domains=sfc)

            >>> print p2_ins
            climlab Process of type <class 'climlab.radiation.insolation.P2Insolation'>.
            State variables and domain shapes:
            The subprocess tree:
            top: <class 'climlab.radiation.insolation.P2Insolation'>

    """
    def __init__(self, S0=const.S0, s2=-0.48, **kwargs):
        super(P2Insolation, self).__init__(S0=S0, **kwargs)
        self.s2 = s2

    @property
    def s2(self):
        """Property of second legendre polynomial factor s2.

        s2 in following equation:

        .. math::

            S(\\varphi) = \\frac{S_0}{4} \\left( 1 + s_2 P_2(x) \\right)

        :getter:    Returns the s2 parameter which is stored in attribute ``self._s2``.
        :setter:    * sets s2 which is addressed as ``self._S0`` to the new value
                    * updates the parameter dictionary ``self.param['s2']`` and
                    * calls method :func:`_compute_fixed`
        :type:      float

        """
        return self._s2
    @s2.setter
    def s2(self, value):
        self._s2 = value
        self.param['s2'] = value
        self._compute_fixed()

    def _compute_fixed(self):
        phi = np.deg2rad(self.lat)
        #  Why is there a silent fail here? Should get rid of this.
        try:
            insolation = self.S0 / 4 * (1. + self.s2 * P2(np.sin(phi)))
            dom = self.domains['default']
            try:
                insolation = to_latlon(insolation, domain=dom)
                self.insolation[:] = insolation
            except:
                self.insolation[:] = Field(insolation, domain=dom)
            # make sure that the diagnostic has the correct field dimensions.
            #self.insolation = Field(insolation, domain=dom)
            self.insolation[:] = Field(insolation, domain=dom)
            self.coszen[:] = self._coszen_from_insolation()
        #  Silent fail only for attribute error: _s2 is not an attribute of self
        #  but s2 parameter is being stored in self._s2
        except AttributeError:
            pass


# These classes calculate insolation based on orbital parameters
#  and astronomical formulas

class AnnualMeanInsolation(_Insolation):
    """A class for latitudewise solar insolation averaged over a year.

    This class computes the solar insolation for each day of the year and
    latitude specified in the domain on the basis of orbital parameters and
    astronomical formulas.

    Therefore it uses the method :func:`~climlab.solar.insolation.daily_insolation`.
    For details how the solar distribution is dependend on orbital parameters
    see there.

    The mean over the year is calculated from data given by
    :func:`~climlab.solar.insolation.daily_insolation` and stored in the
    object's attribute ``self.insolation``

    **Initialization parameters** \n

    :param float S0:    solar constant                                       \n
                        - unit: :math:`\\frac{\\textrm{W}}{\\textrm{m}^2}`   \n
                        - default value: ``1365.2``

    :param dict orb:    a dictionary with three orbital parameters (as provided by
                        :class:`~climlab.solar.orbital.OrbitalTable`):

                        * ``'ecc'`` - eccentricity

                            * unit: dimensionless
                            * default value: ``0.017236``

                        * ``'long_peri'`` - longitude of perihelion (precession angle)

                            * unit: degrees
                            * default value: ``281.37``

                        * ``'obliquity'`` - obliquity angle

                            * unit: degrees
                            * default value: ``23.446``

    **Object attributes** \n

    Additional to the parent class :class:`~climlab.radiation.insolation._Insolation`
    following object attributes are generated and updated during initialization:

    :ivar insolation:       Current insolation in W/m2
    :vartype insolation:    Field

    :ivar coszen:           Cosine of the current solar zenith angle
    :vartype coszen:    Field

    :ivar dict orb:         initialized with given argument ``orb``

    :Example:

        Create regular EBM and replace standard insolation subprocess by
        :class:`~climlab.radation.AnnualMeanInsolation`::

            >>> import climlab
            >>> from climlab.radiation import AnnualMeanInsolation

            >>> # model creation
            >>> model = climlab.EBM()

            >>> print model

        .. code-block:: none
            :emphasize-lines: 12

            climlab Process of type <class 'climlab.model.ebm.EBM'>.
            State variables and domain shapes:
              Ts: (90, 1)
            The subprocess tree:
            top: <class 'climlab.model.ebm.EBM'>
               diffusion: <class 'climlab.dynamics.diffusion.MeridionalDiffusion'>
               LW: <class 'climlab.radiation.AplusBT.AplusBT'>
               albedo: <class 'climlab.surface.albedo.StepFunctionAlbedo'>
                  iceline: <class 'climlab.surface.albedo.Iceline'>
                  cold_albedo: <class 'climlab.surface.albedo.ConstantAlbedo'>
                  warm_albedo: <class 'climlab.surface.albedo.P2Albedo'>
               insolation: <class 'climlab.radiation.insolation.P2Insolation'>

        ::

            >>> # catch model domain for subprocess creation
            >>> sfc = model.domains['Ts']

            >>> # create AnnualMeanInsolation subprocess
            >>> new_insol = AnnualMeanInsolation(domains=sfc, **model.param)

            >>> # add it to the model
            >>> model.add_subprocess('insolation',new_insol)

            >>> print model

        .. code-block:: none
            :emphasize-lines: 12

            climlab Process of type <class 'climlab.model.ebm.EBM'>.
            State variables and domain shapes:
              Ts: (90, 1)
            The subprocess tree:
            top: <class 'climlab.model.ebm.EBM'>
               diffusion: <class 'climlab.dynamics.diffusion.MeridionalDiffusion'>
               LW: <class 'climlab.radiation.AplusBT.AplusBT'>
               albedo: <class 'climlab.surface.albedo.StepFunctionAlbedo'>
                  iceline: <class 'climlab.surface.albedo.Iceline'>
                  cold_albedo: <class 'climlab.surface.albedo.ConstantAlbedo'>
                  warm_albedo: <class 'climlab.surface.albedo.P2Albedo'>
               insolation: <class 'climlab.radiation.insolation.AnnualMeanInsolation'>

    """
    def __init__(self, S0=const.S0, orb=const.orb_present, **kwargs):
        super(AnnualMeanInsolation, self).__init__(S0=S0, **kwargs)
        #self.param['orb'] = orb
        self.orb = orb
        self._compute_fixed()

    @property
    def orb(self):
        """Property of dictionary for orbital parameters.

        orb contains: (for more information see :class:`~climlab.solar.orbital.OrbitalTable`)

        * ``'ecc'`` - eccentricity [unit: dimensionless]
        * ``'long_peri'`` - longitude of perihelion (precession angle) [unit: degrees]
        * ``'obliquity'`` - obliquity angle [unit: degrees]

        :getter:    Returns the orbital dictionary which is stored in attribute
                    ``self._orb``.
        :setter:    * sets orb which is addressed as ``self._orb`` to the new value
                    * updates the parameter dictionary ``self.param['orb']`` and
                    * calls method :func:`_compute_fixed`
        :type:      dict

        """
        return self._orb
    @orb.setter
    def orb(self, value):
        self._orb = value
        self.param['orb'] = value
        self._compute_fixed()

    def _daily_insolation_array(self):
        lat = self.lat
        days_of_year = self.time['days_of_year']
        orb = self.orb
        S0 = self.S0
        return daily_insolation(lat, days_of_year, orb=orb, S0=S0)

    def _compute_fixed(self):
        try:
            temp_array = self._daily_insolation_array()
            insolation = np.mean(temp_array, axis=1)
            # make sure that the diagnostic has the correct field dimensions.
            dom = self.domains['default']
            try:
                insolation = to_latlon(insolation, domain=dom)
                self.insolation[:] = insolation
            except:
                self.insolation[:] = Field(insolation, domain=dom)
            self.coszen[:] = self._coszen_from_insolation()
        #  Silent fail only for attribute error: _orb is not an attribute of self
        #  but orb parameter is being stored in self._orb
        except AttributeError:
            pass


class DailyInsolation(AnnualMeanInsolation):
    """A class to compute latitudewise daily solar insolation for specific
    days of the year.

    This class computes the solar insolation on basis of orbital parameters and
    astronomical formulas.

    Therefore it uses the method :func:`~climlab.solar.insolation.daily_insolation`.
    For details how the solar distribution is dependend on orbital parameters
    see there.

    **Initialization parameters** \n

    :param float S0:    solar constant                              \n
                        - unit: :math:`\\frac{\\textrm{W}}{\\textrm{m}^2}`   \n
                        - default value: ``1365.2``

    :param dict orb:    a dictionary with orbital parameters:

                        * ``'ecc'`` - eccentricity

                            * unit: dimensionless
                            * default value: ``0.017236``

                        * ``'long_peri'`` - longitude of perihelion (precession angle)

                            * unit: degrees
                            * default value: ``281.37``

                        * ``'obliquity'`` - obliquity angle

                            * unit: degrees
                            * default value: ``23.446``

    **Object attributes** \n

    Additional to the parent class :class:`~climlab.radiation.insolation._Insolation`
    following object attributes are generated and updated during initialization:

    :ivar insolation:       Current insolation in W/m2
    :vartype insolation:    Field

    :ivar coszen:           Cosine of the current solar zenith angle
    :vartype coszen:    Field

    :ivar dict orb:         initialized with given argument ``orb``


    :Example:

        Create regular EBM and replace standard insolation subprocess by
        :class:`~climlab.radation.DailyInsolation`::

            >>> import climlab
            >>> from climlab.radiation import DailyInsolation

            >>> # model creation
            >>> model = climlab.EBM()

            >>> print model

        .. code-block:: none
            :emphasize-lines: 12

            climlab Process of type <class 'climlab.model.ebm.EBM'>.
            State variables and domain shapes:
              Ts: (90, 1)
            The subprocess tree:
            top: <class 'climlab.model.ebm.EBM'>
               diffusion: <class 'climlab.dynamics.diffusion.MeridionalDiffusion'>
               LW: <class 'climlab.radiation.AplusBT.AplusBT'>
               albedo: <class 'climlab.surface.albedo.StepFunctionAlbedo'>
                  iceline: <class 'climlab.surface.albedo.Iceline'>
                  cold_albedo: <class 'climlab.surface.albedo.ConstantAlbedo'>
                  warm_albedo: <class 'climlab.surface.albedo.P2Albedo'>
               insolation: <class 'climlab.radiation.insolation.P2Insolation'>

        ::

            >>> # catch model domain for subprocess creation
            >>> sfc = model.domains['Ts']

            >>> # create DailyInsolation subprocess and add it to the model
            >>> model.add_subprocess('insolation',DailyInsolation(domains=sfc, **model.param))

            >>> print model

        .. code-block:: none
            :emphasize-lines: 12

            climlab Process of type <class 'climlab.model.ebm.EBM'>.
            State variables and domain shapes:
              Ts: (90, 1)
            The subprocess tree:
            top: <class 'climlab.model.ebm.EBM'>
               diffusion: <class 'climlab.dynamics.diffusion.MeridionalDiffusion'>
               LW: <class 'climlab.radiation.AplusBT.AplusBT'>
               albedo: <class 'climlab.surface.albedo.StepFunctionAlbedo'>
                  iceline: <class 'climlab.surface.albedo.Iceline'>
                  cold_albedo: <class 'climlab.surface.albedo.ConstantAlbedo'>
                  warm_albedo: <class 'climlab.surface.albedo.P2Albedo'>
               insolation: <class 'climlab.radiation.insolation.DailyInsolation'>

    """

    def _compute_fixed(self):
        try:
            self.insolation_array = self._daily_insolation_array()
        except AttributeError:
            pass

    def _get_current_insolation(self):
        insolation_array = self.insolation_array
        # make sure that the diagnostic has the correct field dimensions.
        dom = self.domains['default']
        time_index = self.time['day_of_year_index']   # THIS ONLY WORKS IF self IS THE MASTER PROCESS
        insolation = insolation_array[..., time_index]
        if 'lon' in dom.axes:
            # insolation is latitude-only, need to broadcast across longitude
            #  assumption is axes are ordered (lat, lon, depth)
            #  NOTE this is a clunky hack and all this will go away
            #  when we use xarray structures for these internals
            insolation = np.tile(insolation[...,np.newaxis], dom.axes['lon'].num_points)
        self.insolation[:] = Field(insolation, domain=dom)
        self.coszen[:] = self._coszen_from_insolation()



import numpy as np

a = 6.373E6      # Radius of Earth (m)
Lhvap = 2.5E6    # Latent heat of vaporization (J / kg)
Lhsub = 2.834E6   # Latent heat of sublimation (J / kg)
Lhfus = Lhsub - Lhvap  # Latent heat of fusion (J / kg)
cp = 1004.     # specific heat at constant pressure for dry air (J / kg / K)
Rd = 287.         # gas constant for dry air (J / kg / K)
kappa = Rd / cp
Rv = 461.5       # gas constant for water vapor (J / kg / K)
cpv = 1875.   # specific heat at constant pressure for water vapor (J / kg / K)
Omega = 2 * np.math.pi / 24. / 3600.  # Earth's rotation rate, (s**(-1))
g = 9.8          # gravitational acceleration (m / s**2)
kBoltzmann = 1.3806488E-23  # the Boltzmann constant (J / K)
c_light = 2.99792458E8   # speed of light (m/s)
hPlanck = 6.62606957E-34  # Planck's constant (J s)
# sigma = 5.67E-8  # Stefan-Boltzmann constant (W / m**2 / K**4)
#  sigma derived from fundamental constants
sigma = (2*np.pi**5 * kBoltzmann**4) / (15 * c_light**2 * hPlanck**3)

S0 = 1365.2       # solar constant (W / m**2)
# value is consistent with Trenberth and Fasullo, Surveys of Geophysics 2012

ps = 1000.       # approximate surface pressure (mb or hPa)

rho_w = 1000.    # density of water (kg / m**3)
cw = 4181.3      # specific heat of liquid water (J / kg / K)

tempCtoK = 273.15   # 0degC in Kelvin
tempKtoC = -tempCtoK  # 0 K in degC
mb_to_Pa = 100.  # conversion factor from mb to Pa

#  Some useful time conversion factors
seconds_per_minute = 60.
minutes_per_hour = 60.
hours_per_day = 24.

# the length of the "tropical year" -- time between vernal equinoxes
# This value is consistent with Berger (1978)
# "Long-Term Variations of Daily Insolation and Quaternary Climatic Changes"
days_per_year = 365.2422
seconds_per_hour = minutes_per_hour * seconds_per_minute
minutes_per_day = hours_per_day * minutes_per_hour
seconds_per_day = hours_per_day * seconds_per_hour
seconds_per_year = seconds_per_day * days_per_year
minutes_per_year = seconds_per_year / seconds_per_minute
hours_per_year = seconds_per_year / seconds_per_hour
#  average lenghts of months based on dividing the year into 12 equal parts
months_per_year = 12.
seconds_per_month = seconds_per_year / months_per_year
minutes_per_month = minutes_per_year / months_per_year
hours_per_month = hours_per_year / months_per_year
days_per_month = days_per_year / months_per_year

area_earth = 4 * np.math.pi * a**2

# present-day orbital parameters, in the same format generated by orbital.py
orb_present = {'ecc': 0.017236, 'long_peri': 281.37, 'obliquity': 23.446}



'''
This chapter documents the source code of the ``climlab`` package.
The focus is on the methods and functions that the user invokes
while using the package.

Nevertheless also the underlying code of the ``climlab`` architecture
has been documented for a comprehensive understanding and traceability.
'''

__version__ = '0.8.0'

# this should ensure that we can still import constants.py as climlab.constants
from .utils import constants, thermo, legendre
# some more useful shorcuts
from .model.column import GreyRadiationModel, RadiativeConvectiveModel, BandRCModel
from .model.ebm import EBM, EBM_annual, EBM_seasonal
from .domain.field import Field, global_mean
from .domain.axis import Axis
from .domain.initial import column_state, surface_state
from .process import Process, TimeDependentProcess, ImplicitProcess, DiagnosticProcess, EnergyBudget
from .process import process_like, get_axes, couple
from .domain.xarray import to_xarray



from __future__ import division
import numpy as np
import climlab
import pytest

@pytest.mark.compiled
@pytest.fixture()
def rcm():
    # initial state (temperatures)
    state = climlab.column_state(num_lev=40, num_lat=1, water_depth=5.)
    ## Create individual physical process models:
    #  fixed relative humidity
    h2o = climlab.radiation.ManabeWaterVapor(state=state, name='H2O')
    #  Hard convective adjustment
    convadj = climlab.convection.ConvectiveAdjustment(state=state, name='ConvectiveAdjustment',
                                                      adj_lapse_rate=6.5)
    # RRTMG radiation with default parameters and interactive water vapor
    rad = climlab.radiation.RRTMG(state=state, albedo=0.2, specific_humidity=h2o.q, name='Radiation')
    # Couple the models
    rcm = climlab.couple([h2o,convadj,rad], name='RCM')
    return rcm

@pytest.mark.compiled
@pytest.mark.fast
def test_convective_adjustment(rcm):
    rcm.step_forward()
    #  test non-scalar critical lapse rate
    num_lev = rcm.lev.size
    rcm.subprocess['ConvectiveAdjustment'].adj_lapse_rate = np.linspace(5., 8., num_lev+1)
    rcm.step_forward()
    #  Test two flags for dry adiabatic adjustment
    rcm.subprocess['ConvectiveAdjustment'].adj_lapse_rate = 'DALR'
    rcm.step_forward()
    rcm.subprocess['ConvectiveAdjustment'].adj_lapse_rate = 'dry adiabat'
    rcm.step_forward()
    #  test pseudoadiabatic critical lapse rate
    rcm.subprocess['ConvectiveAdjustment'].adj_lapse_rate = 'pseudoadiabat'
    rcm.step_forward()
    rcm.subprocess['ConvectiveAdjustment'].adj_lapse_rate = 'MALR'
    rcm.step_forward()
    rcm.subprocess['ConvectiveAdjustment'].adj_lapse_rate = 'moist adiabat'
    rcm.step_forward()

@pytest.mark.compiled
@pytest.mark.fast
def test_coupled_rcm(rcm):
    deltat = rcm.timestep
    ocean_bounds = np.arange(0., 2010., 100.)
    depthax = climlab.Axis(axis_type='depth', bounds=ocean_bounds)
    ocean = climlab.domain.domain.Ocean(axes=depthax)
    ocean_diff = 5.E-4
    Tinitial_ocean = rcm.Ts * np.ones(ocean.shape)
    Tocean = climlab.Field(Tinitial_ocean.copy(), domain=ocean)
    Tatm = rcm.Tatm
    #  Surface temperature Ts is the upper-most grid box of the ocean
    Ts = Tocean[0:1]
    atm_state = {'Tatm': Tatm, 'Ts': Ts}
    rad = climlab.radiation.RRTMG(name='Radiation',
                                  state=atm_state,
                                  specific_humidity=rcm.specific_humidity,
                                  timestep = deltat,
                                  albedo = 0.25,
                                 )
    conv = climlab.convection.ConvectiveAdjustment(name='Convection',
                                                   state=atm_state,
                                                   adj_lapse_rate=6.5,
                                                   timestep=deltat,)

    model = rad + conv
    model.set_state('Tocean', Tocean)
    diff = climlab.dynamics.Diffusion(state={'Tocean': Tocean},
                                K=ocean_diff,
                                diffusion_axis='depth',
                                timestep=deltat * 10,)
    model.add_subprocess('Ocean Heat Uptake', diff)
    for i in range(10):
        model.step_forward()

@pytest.mark.fast
def test_convective_adjustment_multidim():
    #  can we do convective adjustment on a multidimensional grid?
    num_lev = 3
    state = climlab.column_state(num_lev=num_lev, num_lat=2)
    conv = climlab.convection.ConvectiveAdjustment(state=state)
    #  test non-scalar critical lapse rate
    conv.adj_lapse_rate = np.linspace(5., 8., num_lev+1)
    conv.step_forward()
    #  Test two flags for dry adiabatic adjustment
    conv.adj_lapse_rate = 'DALR'
    conv.step_forward()
    conv.adj_lapse_rate = 'dry adiabat'
    conv.step_forward()
    #  test pseudoadiabatic critical lapse rate
    conv.adj_lapse_rate = 'pseudoadiabat'
    conv.step_forward()
    conv.adj_lapse_rate = 'MALR'
    conv.step_forward()
    conv.adj_lapse_rate = 'moist adiabat'
    conv.step_forward()



'''Processes for surface turbulent heat and moisture fluxes

:class:`~climlab.surface.SensibleHeatFlux` and
:class:`~climlab.surface.LatentHeatFlux` implement standard bulk formulae
for the turbulent heat fluxes, assuming that the heating or moistening
occurs in the lowest atmospheric model level.

        :Example:

            Here is an example of setting up a single-column
            Radiative-Convective model with interactive water vapor
            and surface latent and sensible heat fluxes.

            This example also demonstrates *asynchronous coupling*:
            the radiation uses a longer timestep than the other model components::

                import numpy as np
                import climlab
                from climlab import constants as const
                # Temperatures in a single column
                full_state = climlab.column_state(num_lev=30, water_depth=2.5)
                temperature_state = {'Tatm':full_state.Tatm,'Ts':full_state.Ts}
                #  Initialize a nearly dry column (small background stratospheric humidity)
                q = np.ones_like(full_state.Tatm) * 5.E-6
                #  Add specific_humidity to the state dictionary
                full_state['q'] = q
                #  ASYNCHRONOUS COUPLING -- the radiation uses a much longer timestep
                #  The top-level model
                model = climlab.TimeDependentProcess(state=full_state,
                                              timestep=const.seconds_per_hour)
                #  Radiation coupled to water vapor
                rad = climlab.radiation.RRTMG(state=temperature_state,
                                              specific_humidity=full_state.q,
                                              albedo=0.3,
                                              timestep=const.seconds_per_day
                                              )
                #  Convection scheme -- water vapor is a state variable
                conv = climlab.convection.EmanuelConvection(state=full_state,
                                              timestep=const.seconds_per_hour)
                #  Surface heat flux processes
                shf = climlab.surface.SensibleHeatFlux(state=temperature_state, Cd=0.5E-3,
                                              timestep=const.seconds_per_hour)
                lhf = climlab.surface.LatentHeatFlux(state=full_state, Cd=0.5E-3,
                                              timestep=const.seconds_per_hour)
                #  Couple all the submodels together
                model.add_subprocess('Radiation', rad)
                model.add_subprocess('Convection', conv)
                model.add_subprocess('SHF', shf)
                model.add_subprocess('LHF', lhf)
                print(model)

                #  Run the model
                model.integrate_years(1)
                #  Check for energy balance
                print(model.ASR - model.OLR)
'''
from __future__ import division
import numpy as np
from climlab.utils.thermo import qsat
from climlab import constants as const
from climlab.process.energy_budget import EnergyBudget
from climlab.domain.field import Field


class _SurfaceFlux(EnergyBudget):
    '''Abstract parent class for SensibleHeatFlux and LatentHeatFlux'''
    def __init__(self, Cd=3E-3, resistance=1., **kwargs):
        super(_SurfaceFlux, self).__init__(**kwargs)
        self.Cd = Cd
        self.add_input('resistance', resistance)
        self.heating_rate['Tatm'] = np.zeros_like(self.Tatm)
        #  fixed wind speed (for now)
        self.add_input('U', 5. * np.ones_like(self.Ts))
        #  retrieving surface pressure from model grid
        self.ps = self.lev_bounds[-1]

    def _compute_heating_rates(self):
        '''Compute energy flux convergences to get heating rates in :math:`W/m^2`.'''
        self._compute_flux()
        self.heating_rate['Ts'] = -self._flux
        # Modify only the lowest model level
        self.heating_rate['Tatm'][..., -1, np.newaxis] = self._flux

    def _air_density(self, Ta):
        return self.ps * const.mb_to_Pa / const.Rd / Ta


class SensibleHeatFlux(_SurfaceFlux):
    r'''Surface turbulent sensible heat flux implemented through a bulk aerodynamic formula.

    The flux is computed from

    .. math::

        SH = r ~ c_p ~\rho ~ C_D ~ U \left( T_s - T_a \right)


    where:

    - :math:`c_p` and :math:`\rho` are the specific heat and density of air
    - :math:`C_D` is a drag coefficient (stored as ``self.Cd``, default value is 3E-3)
    - :math:`U` is the near-surface wind speed, stored as ``self.U``, default value is 5 m/s
    - :math:`r` is an optional resistance parameter (stored as ``self.resistance``, default value = 1)

    The surface temperature :math:`T_s` is taken directly from ``self.state['Ts']``,
    while the near-surface air temperature :math:`T_a` is taken as the lowest model
    level in ``self.state['Tatm']``

    Diagnostic quantity ``self.SHF`` gives the sensible heat flux in W/m2.

    Temperature tendencies associated with this flux are computed for
    ``Ts`` and for the lowest model level in ``Tatm``. All other tendencies
    (including air temperature tendencies at other levels) are set to zero.
    '''
    def __init__(self, Cd=3E-3, **kwargs):
        super(SensibleHeatFlux, self).__init__(Cd=Cd, **kwargs)
        self.add_diagnostic('SHF', 0.*self.Ts)

    def _compute_flux(self):
        # this ensure same dimensions as Ts
        #  (and use only the lowest model level)
        Ta = Field(self.Tatm[..., -1, np.newaxis], domain=self.Ts.domain)
        Ts = self.Ts
        DeltaT = Ts - Ta
        rho = self._air_density(Ta)
        #  flux from bulk formula
        self._flux = self.resistance * const.cp * rho * self.Cd * self.U * DeltaT
        self.SHF = self._flux



class LatentHeatFlux(_SurfaceFlux):
    r'''Surface turbulent latent heat flux implemented through a bulk aerodynamic formula.

    The flux is computed from

    .. math::

        LH = r ~ L ~\rho ~ C_D ~ U \left( q_s - q_a \right)


    where:

    - :math:`L` and :math:`\rho` are the latent heat of vaporization and density of air
    - :math:`C_D` is a drag coefficient (stored as ``self.Cd``, default value is 3E-3)
    - :math:`U` is the near-surface wind speed, stored as ``self.U``, default value is 5 m/s
    - :math:`r` is an optional resistance parameter (stored as ``self.resistance``, default value = 1)

    The surface specific humidity :math:`q_s` is computed as the saturation specific
    humidity at the surface temperature ``self.state['Ts']`` and surface pressure
    ``self.ps``, while the near-surface specific humidity :math:`q_a` is taken as the lowest model
    level in the field ``self.q`` (which must be provided either as a state variable or as input).

    Two diagnostics are computed:

    - ``self.LHF`` gives the sensible heat flux in W/m2.
    - ``self.evaporation`` gives the evaporation rate in kg/m2/s (or mm/s)

    How the tendencies are computed depends on whether specific humidity ``q``
    is a state variable (i.e. is present in ``self.state``):

    - If ``q`` is in ``self.state`` then the evaporation determines the specific humidity tendency ``self.tendencies['q']``. The water vapor is added to the lowest model level only. Evaporation cools the surface through the surface tendency ``self.tendencies['Ts']``. Air temperature tendencies are zero everywhere.
    - If ``q`` is not in ``self.state`` then we compute an equivalent air temperature tendency for the lowest model layer instead of a specific humidity tendency (i.e. the latent heat flux is applied in the same way as a sensible heat flux).

    This process does not apply a tendency to the surface water amount.
    In the absence of other water processes this implies an infinite water source at the surface (slab ocean).
    '''
    def __init__(self, Cd=3E-3, **kwargs):
        super(LatentHeatFlux, self).__init__(Cd=Cd, **kwargs)
        self.add_diagnostic('LHF', 0.*self.Ts)
        self.add_diagnostic('evaporation', 0.*self.Ts)  # in kg/m2/s or mm/s

    def _compute_flux(self):
        #  specific humidity at lowest model level
        #  assumes pressure is the last axis
        q = Field(self.q[..., -1, np.newaxis], domain=self.Ts.domain)
        Ta = Field(self.Tatm[..., -1, np.newaxis], domain=self.Ts.domain)
        qs = qsat(self.Ts, self.ps)
        Deltaq = Field(qs - q, domain=self.Ts.domain)
        rho = self._air_density(Ta)
        #  flux from bulk formula
        self._flux = self.resistance * const.Lhvap * rho * self.Cd * self.U * Deltaq
        self.LHF[:] = self._flux
        # evporation rate, convert from W/m2 to kg/m2/s (or mm/s)
        self.evaporation[:] = self.LHF/const.Lhvap

    def _compute(self):
        '''Overides the _compute method of EnergyBudget'''
        tendencies = self._temperature_tendencies()
        if 'q' in self.state:
            # in a model with active water vapor, this flux should affect
            #  water vapor tendency, NOT air temperature tendency!
            tendencies['Tatm'] *= 0.
            Pa_per_hPa = 100.
            air_mass_per_area = self.Tatm.domain.lev.delta[...,-1] * Pa_per_hPa / const.g
            specific_humidity_tendency = 0.*self.q
            specific_humidity_tendency[...,-1,np.newaxis] = self.LHF/const.Lhvap / air_mass_per_area
            tendencies['q'] = specific_humidity_tendency
        return tendencies



from __future__ import division
from builtins import object
import numpy as np
# experimental...  this is not well documented!
from numpy.core.umath_tests import matrix_multiply

'''
Testing multi-dimensional column radiation

:Example:

    .. code-block:: python

        import numpy as np
        import climlab
        sfc, atm = climlab.domain.zonal_mean_column()
        absorb = np.ones(atm.shape)
        trans = climlab.radiation.transmissivity.Transmissivity(absorptivity=absorb,axis=1)
        fromspace = np.zeros(sfc.shape)
        emission = 200*np.ones(atm.shape)
        A = trans.flux_down(fluxDownTop=fromspace, emission=emission)
        A.shape

'''

class Transmissivity(object):
    '''Class for calculating and store transmissivity between levels,
    and computing radiative fluxes between levels.

    Input: numpy array of absorptivities.
    It is assumed that the last dimension is vertical levels.

    Attributes: (all stored as numpy arrays):

    * N: number of levels
    * absorptivity: level absorptivity (N)
    * transmissivity: level transmissivity (N)
    * Tup: transmissivity matrix for upwelling beam (N+1, N+1)
    * Tdown: transmissivity matrix for downwelling beam (N+1, N+1)


    Example for N = 3 atmospheric layers:

    tau is a vector of transmissivities

    .. math::

        \\tau = \\left[ 1, \\tau_0, \\tau_1, \\tau_2 \\right]

    A is a matrix

    .. math::

        A= \\left[ \\begin{array}{cccc}
        1       & 1         & 1         & 1         \\\\
        \\tau_0 & 1         & 1         & 1         \\\\
        \\tau_0 & \\tau_1   & 1         & 1         \\\\
        \\tau_0 & \\tau_1   & \\tau_2   & 1         \\\\
        \\end{array} \\right]

    We then take the cumulative product along columns,
    and finally take the lower triangle of the result to get

    .. math::

        Tup= \\left[ \\begin{array}{cccc}
                               1 &               0 &       0 &  0   \\\\
                         \\tau_0 &               1 &       0 &  0   \\\\
                 \\tau_1 \\tau_0 &         \\tau_1 &       1 &  0   \\\\
        \\tau_2 \\tau_1 \\tau_00 & \\tau_2 \\tau_1 & \\tau_2 &  1   \\\\
        \\end{array} \\right]

    and Tdown = transpose(Tup)

    Construct an emission vector for the downwelling beam:

    Edown = [E0, E1, E2, fromspace]

    Now we can get the downwelling beam by matrix multiplication:

    D = Tdown * Edown

    For the upwelling beam, we start by adding the reflected part
    at the surface to the surface emissions:

    Eup = [emit_sfc + albedo_sfc*D[0], E0, E1, E2]

    So that the upwelling flux is

    U = Tup * Eup

    The total flux, positive up is thus

    F = U - D

    The absorbed radiation at the surface is then -F[0]
    The absorbed radiation in the atmosphere is the flux convergence:

    -diff(F)

    '''
    #  quick hack to get some simple cloud albedo
    def __init__(self, absorptivity, reflectivity=None, axis=0):
        self.axis = axis
        #if absorptivity.ndim is not 1:
        #    raise ValueError('absorptivity argument must be a vector')
        if reflectivity is None:
            reflectivity = np.zeros_like(absorptivity)
        self.reflectivity = reflectivity
        self.absorptivity = absorptivity
        self.transmissivity = 1 - absorptivity - reflectivity
        self.shape = self.absorptivity.shape
        N = np.size(self.absorptivity, axis=self.axis)
        self.N = N
        #  For now, let's assume that the vertical axis is the last axis
        Tup, Tdown = compute_T_vectorized(self.transmissivity)
        self.Tup = Tup
        self.Tdown = Tdown

    #def flux_down(self, fluxDownTop, emission=None):
    def flux_up(self, fluxUpBottom, emission=None):
        '''Compute downwelling radiative flux at interfaces between layers.

        Inputs:

            * fluxDownTop: flux down at top
            * emission: emission from atmospheric levels (N)
              defaults to zero if not given

        Returns:

            * vector of downwelling radiative flux between levels (N+1)
              element 0 is the flux down to the surface.

        '''
        if emission is None:
            emission = np.zeros_like(self.absorptivity)
        E = np.concatenate((emission, np.atleast_1d(fluxUpBottom)), axis=-1)
        #  dot product (matrix multiplication) along last axes
        return np.squeeze(matrix_multiply(self.Tup, E[..., np.newaxis]))

    def flux_reflected_up(self, fluxDown, albedo_sfc=0.):
        #reflectivity = np.concatenate((np.atleast_1d(albedo_sfc),
        #                               self.reflectivity), axis=-1)
        reflectivity = np.concatenate((self.reflectivity, np.atleast_1d(albedo_sfc)), axis=-1)
        return reflectivity*fluxDown

    #def flux_up(self, fluxUpBottom, emission=None):
    def flux_down(self, fluxDownTop, emission=None):
        '''Compute upwelling radiative flux at interfaces between layers.

        Inputs:

            * fluxUpBottom: flux up from bottom
            * emission: emission from atmospheric levels (N)
              defaults to zero if not given

        Returns:
            * vector of upwelling radiative flux between levels (N+1)
              element N is the flux up to space.

        '''
        if emission is None:
            emission = np.zeros_like(self.absorptivity)
        E = np.concatenate((np.atleast_1d(fluxDownTop),emission), axis=-1)
        #  dot product (matrix multiplication) along last axes
        return np.squeeze(matrix_multiply(self.Tdown, E[..., np.newaxis]))
#
#def compute_T(transmissivity):
#    # fully vectorized version
#    # works on a vector transmissivity
#    N = transmissivity.size
#    tau = np.concatenate((np.atleast_1d(1.), transmissivity))
#    B = np.tile(tau, (N+1,1)).transpose()
#    A = np.tril(B,k=-1) + np.tri(N+1).transpose()
#    Tup = np.tril(np.cumprod(A, axis=0))
#    Tdown = np.transpose(Tup)
#    return Tup, Tdown

def compute_T_vectorized(transmissivity):
    #  really vectorized version... to work with arbitrary dimensions
    #  of input transmissivity
    #  Assumption is the last dimension of transmissivity is vertical
    trans_shape = np.shape(transmissivity)
    N = trans_shape[-1]
    otherdims = trans_shape[:-1]
    ones = np.ones(otherdims)[..., np.newaxis]
    tau = np.concatenate((ones, transmissivity), axis=-1)
    tiletau = np.tile(tau[..., np.newaxis],N+1)
    # equivalent to taking transpose of last two axes
    #B = np.rollaxis(tiletau, -1, -2)
    B = tiletau
    matdims = np.append(np.array(otherdims),[1,1])
    # dimensions of matrix should be [otherdims,N+1,N+1]
    tri = np.tile(np.tri(N+1).transpose(),matdims)
    #  np.tril refuses to broadcast over other dims in numpy < 1.9
    #  use a custom version instead, below
    #  Performance is BETTER with numpy 1.9
    A = tril(B,k=-1) + tri
    #Tup = tril(np.cumprod(A, axis=-2))
    ##  transpose over last two axes
    #Tdown = np.rollaxis(Tup, -1, -2)
    Tdown = tril(np.cumprod(A, axis=-2))
    #  transpose over last two axes
    Tup = np.rollaxis(Tdown, -1, -2)
    return Tup, Tdown


def tril(array, k=0):
    '''Lower triangle of an array.
    Return a copy of an array with elements above the k-th diagonal zeroed.
    Need a multi-dimensional version here because numpy.tril does not
    broadcast for numpy verison < 1.9.'''
    try:
        tril_array = np.tril(array, k=k)
    except:
        # have to loop
        tril_array = np.zeros_like(array)
        shape = array.shape
        otherdims = shape[:-2]
        for index in np.ndindex(otherdims):
            tril_array[index] = np.tril(array[index], k=k)
    return tril_array



from __future__ import division
import numpy as np
from climlab.process.diagnostic import DiagnosticProcess
from climlab import constants as const
from climlab.utils.thermo import clausius_clapeyron


class FixedRelativeHumidity(DiagnosticProcess):
    def __init__(self, relative_humidity=0.77, qStrat=5.E-6, **kwargs):
        '''Compute water vapor mixing ratio profile
        Assuming constant relative humidity.

        relative_humidity is the specified RH.
        Same value is applied everywhere.
        qStrat is the minimum specific humidity, ensuring that there is
        some water vapor in the stratosphere.

        The attribute RH_profile can be modified to set different
        vertical profiles of relative humidity
        (see daughter class ManabeWaterVapor() ).'''
        super(FixedRelativeHumidity, self).__init__(**kwargs)
        newinput = ['relative_humidity',
                    'qStrat',
                    'RH_profile',]
        self.declare_input(newinput)
        self.relative_humidity = relative_humidity
        self.qStrat = qStrat
        self.RH_profile = self.relative_humidity * np.ones_like(self.Tatm)
        #  go ahead and set the initial q based on initial temperature
        self.add_diagnostic('q', 0.*self.Tatm)
        self._compute()

    def _compute(self):
        es = clausius_clapeyron(self.Tatm)
        e = self.RH_profile * es
        # convert to specific humidity (assume dilute)
        qH2O = e/self.lev * const.Rd / const.Rv
        #  mixing ratio can't be smaller than qStrat
        #  (need some water in the stratosphere!)
        q = np.maximum(self.qStrat, qH2O)
        #  Just set this directly here
        q_adjustment = q - self.q
        self.q += q_adjustment
        return {}



class ManabeWaterVapor(FixedRelativeHumidity):
    def __init__(self, **kwargs):
        '''Compute water vapor mixing ratio profile following
        Manabe and Wetherald JAS 1967
        Fixed surface relative humidity and a specified fractional profile.

        relative_humidity is the specified surface RH
        qStrat is the minimum specific humidity, ensuring that there is
        some water vapor in the stratosphere.'''
        super(ManabeWaterVapor, self).__init__(**kwargs)
        p = self.lev
        Q = p / const.ps
        self.RH_profile = self.relative_humidity * ((Q - 0.02) / (1-0.02))
        self._compute() # call this again so the diagnostic is correct initially



from __future__ import division, print_function
import os


def configuration(parent_package='',top_path=None):
    from numpy.distutils.misc_util import Configuration
    config = Configuration(package_name='orbital', parent_name=parent_package, top_path=top_path)
    config.add_data_files(os.path.join('data','README'))
    return config

if __name__ == '__main__':
    print('This is the wrong setup.py file to run')



import setuptools, os

VERSION = '0.8.0'

# BEFORE importing setuptools, remove MANIFEST. Otherwise it may not be
# properly updated when the contents of directories change (true for distutils,
# not sure about setuptools).
if os.path.exists('MANIFEST'):
    os.remove('MANIFEST')

def readme():
    with open('README.rst') as f:
        return f.read()

setuptools.setup(
    name='climlab',
    version=VERSION,
    description='Package for process-oriented climate modeling',
    long_description=readme(),
    classifiers=[
      'License :: OSI Approved :: MIT License',
      'Programming Language :: Python',
      'Intended Audience :: Education',
      'Intended Audience :: Science/Research',
      'Topic :: Scientific/Engineering :: Atmospheric Science',
    ],
    keywords='climate modeling modelling model ebm radiation radiative-convective earth',
    url='http://github.com/climlab/climlab',
    author='Brian E. J. Rose',
    author_email='brose@albany.edu',
    setup_requires=['numpy'],
    install_requires=['numpy','xarray','scipy'],
    license='MIT',
    packages=setuptools.find_packages(),
)



from __future__ import division
from builtins import next
import numpy as np
from climlab.process.diagnostic import DiagnosticProcess
from climlab.utils.legendre import P2
from climlab.domain.field import Field, global_mean


class ConstantAlbedo(DiagnosticProcess):
    """A class for constant albedo values at all spatial points of the domain.

    **Initialization parameters** \n

    :param float albedo:    albedo values [default: 0.33]

    **Object attributes** \n

    Additional to the parent class
    :class:`~climlab.process.diagnostic.DiagnosticProcess`
    following object attributes are generated and updated during initialization:

    :ivar Field albedo:             attribute to store the albedo value.
                                    During initialization the
                                    :func:`albedo` setter is called.

    :Example:

        Creation of a constant albedo subprocess on basis of an EBM domain::

            >>> import climlab
            >>> from climlab.surface.albedo import ConstantAlbedo

            >>> # model creation
            >>> model = climlab.EBM()

            >>> sfc = model.domains['Ts']

            >>> # subprocess creation
            >>> const_alb = ConstantAlbedo(albedo=0.3, domains=sfc, **model.param)

    """
    def __init__(self, albedo=0.33, **kwargs):
        '''Uniform prescribed albedo.'''
        super(ConstantAlbedo, self).__init__(**kwargs)
        dom = next(iter(self.domains.values()))
        self.add_diagnostic('albedo', Field(albedo, domain=dom))
        #self.albedo = albedo

    # @property
    # def albedo(self):
    #     """Property of albedo value.
    #
    #     :getter:    Returns the albedo value which is stored in diagnostic dict
    #                 ``self.diagnostic['albedo']``
    #     :setter:    * sets albedo which is addressed as ``diagnostics['albedo']``
    #                   to the new value through creating a Field on the basis
    #                   of domain ``self.domain['default']``
    #                 * updates the parameter dictionary ``self.param['albedo']``
    #     :type:      Field
    #
    #     """
    #     return self.diagnostics['albedo']
    # @albedo.setter
    # def albedo(self, value):
    #     #dom = self.domains['default']
    #     #  this is a more robust way to get the single value from dictionary:
    #     dom = self.domains.itervalues().next()
    #     self.diagnostics['albedo'] = Field(value, domain=dom)
    #     self.param['albedo'] = value


class P2Albedo(DiagnosticProcess):
    """A class for parabolic distributed albedo values across the domain
    on basis of the second order Legendre Polynomial.

    Calculates the latitude dependent albedo values as

    .. math::

         \\alpha(\\varphi) = a_0 + a_2 P_2(x)

    where :math:`P_2(x) = \\frac{1}{2} (3x^2 - 1)` is the second order Legendre Polynomial
    and :math:`x=sin(\\varphi)`.

    **Initialization parameters** \n

    :param float a0:    basic parameter for albedo function [default: 0.33]
    :param float a2:    factor for second legendre polynominal term in albedo
                        function [default: 0.25]

    **Object attributes** \n

    Additional to the parent class
    :class:`~climlab.process.diagnostic.DiagnosticProcess`
    following object attributes are generated and updated during initialization:

    :ivar float a0:                 attribute to store the albedo parameter a0.
                                    During initialization the
                                    :func:`a0` setter is called.
    :ivar float a2:                 attribute to store the albedo parameter a2.
                                    During initialization the
                                    :func:`a2` setter is called.
    :ivar dict diagnostics:         key ``'albedo'`` initialized
    :ivar Field albedo:             the subprocess attribute ``self.albedo`` is
                                    created with correct dimensions
                                    (according to ``self.lat``)

    :Example:

        Creation of a parabolic albedo subprocess on basis of an EBM domain::

            >>> import climlab
            >>> from climlab.surface.albedo import P2Albedo

            >>> # model creation
            >>> model = climlab.EBM()

            >>> # modify a0 and a2 values in model parameter dictionary
            >>> model.param['a0']=0.35
            >>> model.param['a2']= 0.10

            >>> # subprocess creation
            >>> p2_alb = P2Albedo(domains=model.domains['Ts'], **model.param)

            >>> p2_alb.a0
            0.33
            >>> p2_alb.a2
            0.1


    """

    def __init__(self, a0=0.33, a2=0.25, **kwargs):
        super(P2Albedo, self).__init__(**kwargs)
        self.a0 = a0
        self.a2 = a2
        self.add_diagnostic('albedo')
        self._compute_fixed()

    @property
    def a0(self):
        """Property of albedo parameter a0.

        :getter:    Returns the albedo parameter value which is stored in attribute
                    ``self._a0``
        :setter:    * sets albedo parameter which is addressed as ``self._a0``
                      to the new value
                    * updates the parameter dictionary ``self.param['a0']``
                    * calls method :func:`_compute_fixed`
        :type:      float

        """
        return self._a0
    @a0.setter
    def a0(self, value):
        self._a0 = value
        self.param['a0'] = value
        self._compute_fixed()
    @property
    def a2(self):
        """Property of albedo parameter a2.

        :getter:    Returns the albedo parameter value which is stored in attribute
                    ``self._a2``
        :setter:    * sets albedo parameter which is addressed as ``self._a2``
                      to the new value
                    * updates the parameter dictionary ``self.param['a2']``
                    * calls method :func:`_compute_fixed`
        :type:      float

        """
        return self._a2
    @a2.setter
    def a2(self, value):
        self._a2 = value
        self.param['a2'] = value
        self._compute_fixed()

    def _compute_fixed(self):
        '''Recompute any fixed quantities after a change in parameters'''
        try:
            lon, lat = np.meshgrid(self.lon, self.lat)
        except:
            lat = self.lat
        phi = np.deg2rad(lat)
        try:
            albedo = self.a0 + self.a2 * P2(np.sin(phi))
        except:
            albedo = np.zeros_like(phi)
        # make sure that the diagnostic has the correct field dimensions.
        #dom = self.domains['default']
        #  this is a more robust way to get the single value from dictionary:
        dom = next(iter(self.domains.values()))
        self.albedo = Field(albedo, domain=dom)



class Iceline(DiagnosticProcess):
    """A class for an Iceline subprocess.

    Depending on a freezing temperature it calculates where on the domain the
    surface is covered with ice, where there is no ice and on which latitude the ice-edge
    is placed.

    **Initialization parameters** \n

    :param float Tf:    freezing temperature where sea water freezes and
                        surface is covered with ice                     \n
                        - unit: :math:`^{\circ} \\textrm{C}`            \n
                        - default value: ``-10``

    **Object attributes** \n

    Additional to the parent class
    :class:`~climlab.process.diagnostic.DiagnosticProcess`
    following object attributes are generated and updated during initialization:

    :ivar dict param:           The parameter dictionary is updated with the
                                input argument ``'Tf'``.
    :ivar dict diagnostics:     keys ``'icelat'`` and ``'ice_area'`` initialized
    :ivar array icelat:         the subprocess attribute ``self.icelat`` is
                                created
    :ivar float ice_area:       the subprocess attribute ``self.ice_area`` is
                                created


    """
    def __init__(self, Tf=-10., **kwargs):
        super(Iceline, self).__init__(**kwargs)
        self.param['Tf'] = Tf
        self.add_diagnostic('icelat')
        self.add_diagnostic('ice_area')
        #  Set diagnostics based on initial conditions
        self.find_icelines()

    def find_icelines(self):
        """Finds iceline according to the surface temperature.

        This method is called by the private function
        :func:`~climlab.surface.albedo.Iceline._compute`
        and updates following attributes according to the freezing temperature
        ``self.param['Tf']`` and the surface temperature ``self.param['Ts']``:

        **Object attributes** \n

        :ivar Field noice:      a Field of booleans which are ``True`` where
                                :math:`T_s \\ge T_f`
        :ivar Field ice:        a Field of booleans which are ``True`` where
                                :math:`T_s < T_f`
        :ivar array icelat:     an array with two elements indicating the
                                ice-edge latitudes
        :ivar float ice_area:   fractional area covered by ice (0 - 1)
        :ivar dict diagnostics: keys ``'icelat'`` and ``'ice_area'`` are updated

        """
        Tf = self.param['Tf']
        Ts = self.state['Ts']
        lat_bounds = self.domains['Ts'].axes['lat'].bounds
        self.noice = np.where(Ts >= Tf, True, False)
        self.ice = np.where(Ts < Tf, True, False)
        #  Ice cover in fractional area
        self.ice_area = global_mean(self.ice * np.ones_like(self.Ts))
        #  Express ice cover in terms of ice edge latitudes
        if self.ice.all():
            # 100% ice cover
            self.icelat = np.array([-0., 0.])
        elif self.noice.all():
            # zero ice cover
            self.icelat = np.array([-90., 90.])
        else:  # there is some ice edge
            # Taking np.diff of a boolean array gives True at the boundaries between True and False
            boundary_indices = np.where(np.diff(self.ice.squeeze()))[0]+1
            # check for asymmetry case: [-90,x] or [x,90]
            #  -> boundary_indices hold only one value for icelat
            if boundary_indices.size == 1:
                if self.ice[0] == True:   # case: [x,90]
                    # extend indice array by missing value for northpole
                    boundary_indices = np.append(boundary_indices, self.ice.size)
                elif  self.ice[-1] == True:   # case: [-90,x]
                    # extend indice array by missing value for northpole
                    boundary_indices = np.insert(boundary_indices,0 ,0)
            # check for asymmetry case: [-90,x] or [x,90]
            #  -> boundary_indices hold only one value for icelat
            if boundary_indices.size == 1:
                if self.ice[0] == True:   # case: [x,90]
                    # extend indice array by missing value for northpole
                    boundary_indices = np.append(boundary_indices, self.ice.size)
                elif  self.ice[-1] == True:   # case: [-90,x]
                    # extend indice array by missing value for northpole
                    boundary_indices = np.insert(boundary_indices,0 ,0)
            self.icelat = lat_bounds[boundary_indices]  # an array of boundary latitudes

    def _compute(self):
        self.find_icelines()
        return {}


class StepFunctionAlbedo(DiagnosticProcess):
    """A step function albedo suprocess.

    This class itself defines three subprocesses that are created during
    initialization:

        * ``'iceline'`` - :class:`Iceline`
        * ``'warm_albedo'`` - :class:`P2Albedo`
        * ``'cold_albedo'`` - :class:`ConstantAlbedo`

    **Initialization parameters** \n

    :param float Tf:    freezing temperature for Iceline subprocess     \n
                        - unit: :math:`^{\circ} \\textrm{C}`            \n
                        - default value: ``-10``
    :param float a0:    basic parameter for P2Albedo subprocess [default: 0.3]
    :param float a2:    factor for second legendre polynominal term in P2Albedo
                        subprocess [default: 0.078]
    :param float ai:    ice albedo value for ConstantAlbedo subprocess
                        [default: 0.62]

    Additional to the parent class
    :class:`~climlab.process.diagnostic.DiagnosticProcess`
    following object attributes are generated/updated during initialization:

    :ivar dict param:               The parameter dictionary is updated with
                                    a couple of the initatilzation input
                                    arguments, namely ``'Tf'``, ``'a0'``,
                                    ``'a2'`` and ``'ai'``.
    :ivar bool topdown:             is set to ``False`` to call subprocess
                                    compute method first
    :ivar dict diagnostics:         key ``'albedo'`` initialized
    :ivar Field albedo:             the subprocess attribute ``self.albedo`` is
                                    created

    :Example:

        Creation of a step albedo subprocess on basis of an EBM domain::

            >>> import climlab
            >>> from climlab.surface.albedo import StepFunctionAlbedo
            >>>
            >>> model = climlab.EBM(a0=0.29, a2=0.1, ai=0.65, Tf=-2)
            >>>
            >>> step_alb = StepFunctionAlbedo(state= model.state, **model.param)
            >>>
            >>> print step_alb
            climlab Process of type <class 'climlab.surface.albedo.StepFunctionAlbedo'>.
            State variables and domain shapes:
              Ts: (90, 1)
            The subprocess tree:
            top: <class 'climlab.surface.albedo.StepFunctionAlbedo'>
               iceline: <class 'climlab.surface.albedo.Iceline'>
               cold_albedo: <class 'climlab.surface.albedo.ConstantAlbedo'>
               warm_albedo: <class 'climlab.surface.albedo.P2Albedo'>

    """
    def __init__(self, Tf=-10., a0=0.3, a2=0.078, ai=0.62, **kwargs):
        super(StepFunctionAlbedo, self).__init__(**kwargs)
        self.param['Tf'] = Tf
        self.param['a0'] = a0
        self.param['a2'] = a2
        self.param['ai'] = ai
        sfc = self.domains['Ts']
        self.add_subprocess('iceline', Iceline(Tf=Tf, state=self.state, timestep=self.timestep))
        warm = P2Albedo(a0=a0, a2=a2, domains=sfc, timestep=self.timestep)
        cold = ConstantAlbedo(albedo=ai, domains=sfc, timestep=self.timestep)
        # remove `albedo` from the diagnostics list for the two subprocesses
        #  because they cause conflicts when passed up the subprocess tree
        for proc in [warm, cold]:
            proc._diag_vars.remove('albedo')
        self.add_subprocess('warm_albedo', warm)
        self.add_subprocess('cold_albedo', cold)
        self.topdown = False  # call subprocess compute methods first
        self.add_diagnostic('albedo', self._get_current_albedo())

    def _get_current_albedo(self):
        '''Simple step-function albedo based on ice line at temperature Tf.'''
        ice = self.subprocess['iceline'].ice
        # noice = self.subprocess['iceline'].diagnostics['noice']
        cold_albedo = self.subprocess['cold_albedo'].albedo
        warm_albedo = self.subprocess['warm_albedo'].albedo
        albedo = Field(np.where(ice, cold_albedo, warm_albedo), domain=self.domains['Ts'])
        return albedo

    def _compute(self):
        self.albedo[:] = self._get_current_albedo()
        return {}



from __future__ import division
from climlab.process.time_dependent_process import TimeDependentProcess


class DiagnosticProcess(TimeDependentProcess):
    """A parent class for all processes that are strictly diagnostic,
    namely that do **not** contribute directly to tendencies of state variables.

    During initialization following attribute is set:

    :ivar time_type:        is set to ``'diagnostic'``
    :vartype time_type:     str

    """
    def __init__(self, **kwargs):
        super(DiagnosticProcess, self).__init__(**kwargs)
        self.time_type = 'diagnostic'

    def _compute(self):
        """Diagnotic process returns no tendencies.

        """
        return {}



r"""CLIMLAB Process objects for advection-diffusion processes of the form

.. math::

    \frac{\partial}{\partial t} \psi(x,t) &= -\frac{1}{w(x)} \frac{\partial}{\partial x} \left[ w(x) ~ \mathcal{F}(x,t) \right] \\
    \mathcal{F} &= U(x) \psi(x) -K(x) ~ \frac{\partial \psi}{\partial x} + F(x)

for a state variable :math:`\psi(x,t)`, diffusivity :math:`K(x)`
in units of :math:`x^2 ~ t^{-1}`, advecting velocity :math:`U(x)`
in units of :math:`x ~ t^{-1}`, and a prescribed flux F(x)
(including boundary conditions) in units of :math:`\psi ~ x ~ t^{-1}`.

The prescribed flux :math:`F(x)` defaults to zero everywhere. The user can
implement a non-zero boundary flux condition by passing a non-zero array
``prescribed_flux`` as input.

:math:`w(x)` is an optional weighting function
for the divergence operator on curvilinear grids.

The diffusivity :math:`K` and velocity :math:`U` can be scalars,
or optionally vectors *specified at grid cell boundaries*
(so their lengths must be exactly 1 greater than the length of :math:`x`).

:math:`K` and :math:`U` can be modified by the user at any time
(e.g., after each timestep, if they depend on other state variables).

A fully implicit timestep is used for computational efficiency. Thus the computed
tendency :math:`\frac{\partial \psi}{\partial t}` will depend on the timestep.

In addition to the tendency over the implicit timestep,
the solver also calculates several diagnostics from the updated state:

- ``diffusive_flux`` given by :math:`-K(x) ~ \frac{\partial \psi}{\partial x}` in units of :math:`[\psi]~[x]`/s
- ``advective_flux`` given by :math:`U(x) \psi(x)` (same units)
- ``total_flux``, the sum of advective, diffusive and prescribed fluxes
- ``flux_convergence`` given by the right hand side of the first equation above, in units of :math:`[\psi]`/s

This base class can be used without modification for diffusion in
Cartesian coordinates (:math:`w=1`). Non-uniformly spaced grids are supported.

The state variable :math:`\psi` may be multi-dimensional, but the diffusion
will operate along a single dimension only.

Other classes implement the weighting for spherical geometry.
"""
from __future__ import division
import numpy as np
from climlab.process.implicit import ImplicitProcess
from climlab.process.process import get_axes
from . import adv_diff_numerics


class AdvectionDiffusion(ImplicitProcess):
    """A parent class for one dimensional implicit advection-diffusion modules.

    **Initialization parameters** \n

    :param float K:                 the diffusivity parameter in units of
                                    :math:`\\frac{[\\textrm{length}]^2}{\\textrm{time}}`
                                    where length is the unit of the spatial axis
                                    on which the diffusion is occuring.
    :param float U:                 Advection velocity in units of
                                    :math:`\\frac{[\\textrm{length}]}{\\textrm{time}}`
    :param str diffusion_axis:      dictionary key for axis on which the
                                    diffusion is occuring in process's domain
                                    axes dictionary
    :param bool use_banded_solver:  input flag, whether to use
                                    :py:func:`scipy.linalg.solve_banded`
                                    instead of :py:func:`numpy.linalg.solve`
                                    [default: False]

    .. note::

        The banded solver :py:func:`scipy.linalg.solve_banded` is faster than
        :py:func:`numpy.linalg.solve` but only works for one dimensional diffusion.

    **Object attributes** \n

    Additional to the parent class
    :class:`~climlab.process.implicit.ImplicitProcess`
    following object attributes are generated or modified during initialization:

    :ivar dict param:               parameter dictionary is extended by
                                    diffusivity parameter K (unit:
                                    :math:`\\frac{[\\textrm{length}]^2}{\\textrm{time}}`)
    :ivar bool use_banded_solver:   input flag specifying numerical solving
                                    method (given during initialization)
    :ivar str diffusion_axis:       dictionary key for axis where diffusion
                                    is occuring:
                                    specified during initialization
                                    or output of method
                                    :func:`_guess_diffusion_axis`
    :ivar array _advdiffTriDiag:        tridiagonal diffusion matrix made by
                                    :func:`_make_diffusion_matrix()` with input
                                    ``self._K_dimensionless``


    :Example:

        Here is an example showing implementation of a vertical diffusion.
        It shows that a subprocess can work on just a subset of the parent process
        state variables.

            .. plot:: code_input_manual/example_diffusion.py
               :include-source:

    """
    def __init__(self,
                 K=0.,
                 U=0.,
                 diffusion_axis=None,
                 use_banded_solver=False,
                 prescribed_flux=0.,
                 **kwargs):
        super(AdvectionDiffusion, self).__init__(**kwargs)
        self.use_banded_solver = use_banded_solver
        if diffusion_axis is None:   # diffusion axis is also advection axis!
            self.diffusion_axis = _guess_diffusion_axis(self)
        else:
            self.diffusion_axis = diffusion_axis
        for dom in list(self.domains.values()):
            points = dom.axes[self.diffusion_axis].points
            bounds = dom.axes[self.diffusion_axis].bounds
        self.diffusion_axis_index = dom.axis_index[self.diffusion_axis]
        #  Cell bounds and centers in length units for diffusion operator
        #  Ensure they have shame dimensions as state var
        for varname, value in self.state.items():
            arr = np.moveaxis(0.*value, self.diffusion_axis_index, -1)
            J = arr.shape[-1]
            sizeJ = tuple([n for n in arr.shape[:-1]] + [J])
            sizeJplus1 = tuple([n for n in arr.shape[:-1]] + [J+1])
            arr[...,:] = points
            self._Xcenter = arr
            self._Xbounds = np.zeros(sizeJplus1)
            self._Xbounds[...,:] = bounds
        self._weight_bounds = np.ones_like(self._Xbounds)  # weights for curvilinear grids
        self._weight_center = np.ones_like(self._Xcenter)
        self.prescribed_flux = prescribed_flux  # flux including boundary conditions
        self.K = K  # Diffusivity in units of [length]**2 / [time]
        self.U = U  # Advecting velocity in units of [length] / [time]
        self.add_diagnostic('diffusive_flux',
            np.moveaxis(0.*self.K*self._weight_bounds,-1,self.diffusion_axis_index))
        self.add_diagnostic('advective_flux', 0.*self.diffusive_flux)
        self.add_diagnostic('total_flux', 0.*self.diffusive_flux)
        for varname, value in self.state.items():
            self.add_diagnostic('flux_convergence',
                np.moveaxis(0.*self._weight_center,-1,self.diffusion_axis_index))

    @property
    def K(self):
        return self._K
    @K.setter  # currently this assumes that Kvalue is scalar or has the right dimensions...
    def K(self, Kvalue):
        self._K = Kvalue
        self._compute_advdiff_matrix()

    @property
    def U(self):
        return self._U
    @U.setter
    def U(self, Uvalue):
        self._U = Uvalue
        self._compute_advdiff_matrix()

    @property
    def prescribed_flux(self):
        return self._prescribed_flux
    @prescribed_flux.setter
    def prescribed_flux(self, fluxvalue):
        self._prescribed_flux = fluxvalue
        for varname, value in self.state.items():
            field = np.moveaxis(value, self.diffusion_axis_index,-1)
        fluxarray = np.ones_like(self._Xbounds) * self._prescribed_flux
        self._source = adv_diff_numerics.compute_source(X=self._Xcenter,
            Xb=self._Xbounds, prescribed_flux=fluxarray,
            prescribed_source=0.*field,
            W=self._weight_center, Wb=self._weight_bounds)

    def _compute_advdiff_matrix(self):
        Karray = np.ones_like(self._Xbounds) * self.K
        try:
            Uarray = np.ones_like(self._Xbounds) * self.U
        except Exception:
            Uarray = 0.*Karray
        self._advdiffTriDiag = adv_diff_numerics.advdiff_tridiag(X=self._Xcenter,
            Xb=self._Xbounds, K=Karray, U=Uarray, W=self._weight_center, Wb=self._weight_bounds,
            use_banded_solver=self.use_banded_solver)

    def _implicit_solver(self):
        newstate = {}
        for varname, value in self.state.items():
            field = np.moveaxis(value, self.diffusion_axis_index,-1)
            result = adv_diff_numerics.implicit_step_forward(field,
                        self._advdiffTriDiag, self._source, self.timestep,
                        use_banded_solver=self.use_banded_solver)
            newstate[varname] = np.moveaxis(result,-1,self.diffusion_axis_index)
        return newstate

    def _update_diagnostics(self, newstate):
        Karray = np.ones_like(self._Xbounds) * self.K
        Uarray = np.ones_like(self._Xbounds) * self.U
        for varname, value in newstate.items():
            field = np.moveaxis(value, self.diffusion_axis_index,-1)
            diff_flux = adv_diff_numerics.diffusive_flux(self._Xcenter,
                                        self._Xbounds, Karray, field)
            adv_flux = adv_diff_numerics.advective_flux(self._Xcenter,
                                        self._Xbounds, Uarray, field)
            self.diffusive_flux[:] = np.moveaxis(diff_flux,-1,self.diffusion_axis_index)
            self.advective_flux[:] = np.moveaxis(adv_flux,-1,self.diffusion_axis_index)
            source = 0.*field
            convergence = adv_diff_numerics.compute_tendency(field,
                self._advdiffTriDiag, source, use_banded_solver=self.use_banded_solver)
            self.flux_convergence[:] = np.moveaxis(convergence,-1,self.diffusion_axis_index)


class Diffusion(AdvectionDiffusion):
    '''1D diffusion only, with advection set to zero.

    Otherwise identical to the parent class AdvectionDiffusion.
    '''
    def __init__(self,
                 K=None,
                 diffusion_axis=None,
                 use_banded_solver=False,
                 **kwargs):
        super(Diffusion, self).__init__(K=K, U=0.,
            diffusion_axis=diffusion_axis,
            use_banded_solver=use_banded_solver, **kwargs)


def _guess_diffusion_axis(process_or_domain):
    """Scans given process, domain or dictionary of domains for a diffusion axis
    and returns appropriate name.

    In case only one axis with length > 1 in the process or set of domains
    exists, the name of that axis is returned. Otherwise an error is raised.

    :param process_or_domain:   input from where diffusion axis should be guessed
    :type process_or_domain:    :class:`~climlab.process.process.Process`,
                                :class:`~climlab.domain.domain._Domain` or
                                :py:class:`dict` of domains
    :raises: :exc:`ValueError` if more than one diffusion axis is possible.
    :returns:                   name of the diffusion axis
    :rtype:                     str

    """
    axes = get_axes(process_or_domain)
    diff_ax = {}
    for axname, ax in axes.items():
        if ax.num_points > 1:
            diff_ax.update({axname: ax})
    if len(list(diff_ax.keys())) == 1:
        return list(diff_ax.keys())[0]
    else:
        raise ValueError('More than one possible diffusion axis.')



from __future__ import division

def walk_processes(top, topname='top', topdown=True, ignoreFlag=False):
    """Generator for recursive tree of climlab processes

    Starts walking from climlab process ``top`` and generates a complete
    list of all processes and sub-processes that are managed from ``top`` process.
    ``level`` indicades the rank of specific process in the process hierarchy:

    .. note::

        * level 0: ``top`` process
            * level 1: sub-processes of ``top`` process
                * level 2: sub-sub-processes of ``top`` process (=subprocesses of level 1 processes)



    The method is based on os.walk().



    :param top:             top process from where walking should start
    :type top:              :class:`~climlab.process.process.Process`
    :param str topname:     name of top process [default: 'top']
    :param bool topdown:    whether geneterate *process_types* in regular or
                            in reverse order [default: True]
    :param bool ignoreFlag: whether ``topdown`` flag should be ignored or not
                            [default: False]
    :returns: name (str), proc (process), level (int)


    :Example:

        ::

            >>> import climlab
            >>> from climlab.utils import walk

            >>> model = climlab.EBM()

            >>> for name, proc, top_proc in walk.walk_processes(model):
            ...     print name
            ...
            top
            diffusion
            LW
            iceline
            cold_albedo
            warm_albedo
            albedo
            insolation

    """
    if not ignoreFlag:
        flag = topdown
    else:
        flag = True
    proc = top
    level = 0
    if flag:
        yield topname, proc, level
    if len(proc.subprocess) > 0:  # there are sub-processes
        level += 1
        for name, subproc in proc.subprocess.items():
            for name2, subproc2, level2 in walk_processes(subproc,
                                                          topname=name,
                                                          topdown=subproc.topdown,
                                                          ignoreFlag=ignoreFlag):
                yield name2, subproc2, level+level2
    if not flag:
        yield topname, proc, level


def process_tree(top, name='top'):
    """Creates a string representation of the process tree for process top.

    This method uses the :func:`walk_processes` method to create the process tree.

    :param  top:            top process for which process tree string should be
                            created
    :type top:              :class:`~climlab.process.process.Process`
    :param str name:        name of top process
    :returns:               string representation of the process tree
    :rtype:                 str

    :Example:

        ::

            >>> import climlab
            >>> from climlab.utils import walk

            >>> model = climlab.EBM()
            >>> proc_tree_str = walk.process_tree(model, name='model')

            >>> print proc_tree_str
            model: <class 'climlab.model.ebm.EBM'>
               diffusion: <class 'climlab.dynamics.diffusion.MeridionalDiffusion'>
               LW: <class 'climlab.radiation.AplusBT.AplusBT'>
               albedo: <class 'climlab.surface.albedo.StepFunctionAlbedo'>
                  iceline: <class 'climlab.surface.albedo.Iceline'>
                  cold_albedo: <class 'climlab.surface.albedo.ConstantAlbedo'>
                  warm_albedo: <class 'climlab.surface.albedo.P2Albedo'>
               insolation: <class 'climlab.radiation.insolation.P2Insolation'>

    """
    str1 = ''
    for name, proc, level in walk_processes(top, name, ignoreFlag=True):
        indent = ' ' * 3 * (level)
        str1 += ('{}{}: {}\n'.format(indent, name, type(proc)))
    return str1



from __future__ import division
import numpy as np
from climlab import thermo
from climlab.process.energy_budget import EnergyBudget


class SurfaceRadiation(EnergyBudget):
    def __init__(self, albedo_sfc=None, **kwargs):
        super(SurfaceRadiation, self).__init__(**kwargs)
        newinput = ['albedo_sfc',
                    'LW_from_atm',
                    'SW_from_atm',]
        self.declare_input(newinput)
        if albedo_sfc is None:
            self.albedo_sfc = np.zeros_like(self.Ts)
        else:
            self.albedo_sfc = albedo_sfc*np.ones_like(self.Ts)
        self.LW_from_atm = 0. * self.Ts
        self.SW_from_atm = 0. * self.Ts
        newdiags = ['LW_to_atm',
                    'SW_to_atm']
        for name in newdiags:
            self.add_diagnostic(name)
        self.LW_to_atm = 0. * self.Ts
        self.SW_to_atm = 0. * self.Ts
        self.heating_rate['Tatm'] = np.zeros_like(self.Tatm)

    def _compute_emission(self):
        return thermo.blackbody_emission(self.Ts)

    def _compute_reflected_flux(self):
        return self.SW_from_atm * self.albedo_sfc

    def _compute_heating_rates(self):
        '''Compute energy flux convergences to get heating rates in :math:`W/m^2`.'''
        self.LW_to_atm = self._compute_emission()
        self.SW_to_atm = self._compute_reflected_flux()
        self.heating_rate['Ts'] = ( self.LW_from_atm - self.LW_to_atm
                                  + self.SW_from_atm - self.SW_to_atm )



import climlab
from climlab.dynamics.diffusion import Diffusion
import matplotlib.pyplot as plt

c = climlab.GreyRadiationModel()
K = 0.5
d = Diffusion(K=K, state = {'Tatm':c.state['Tatm']}, **c.param)

c.add_subprocess('diffusion',d)

### Integrate & Plot ###

fig = plt.figure( figsize=(6,4))
ax = fig.add_subplot(111)

ax.plot(c.lev, c.state['Tatm'], label='step 0')
c.step_forward()
ax.plot(c.lev, c.state['Tatm'], label='step 1')

ax.invert_xaxis()
ax.set_title('Diffusion subprocess')
ax.set_xlabel('level (mb)')
#ax.set_xticks([])
ax.set_ylabel('temperature (K)')
ax.legend(loc='best')
plt.show()



from __future__ import division, absolute_import
from builtins import str
from builtins import object
from xarray import Dataset, DataArray
import warnings


def Field_to_xarray(field):
    '''Convert a climlab.Field object to xarray.DataArray'''
    dom = field.domain
    dims = []; dimlist = []; coords = {};
    for axname in dom.axes:
        dimlist.append(axname)
        try:
            assert field.interfaces[dom.axis_index[axname]]
            bounds_name = axname + '_bounds'
            dims.append(bounds_name)
            coords[bounds_name] = dom.axes[axname].bounds
        except:
            dims.append(axname)
            coords[axname] = dom.axes[axname].points
    #  Might need to reorder the data
    da = DataArray(field.transpose([dom.axis_index[name] for name in dimlist]),
                      dims=dims, coords=coords)
    for name in dims:
        try:
            da[name].attrs['units'] = dom.axes[name].units
        except:
            pass
    return da

def state_to_xarray(state):
    '''Convert a dictionary of climlab.Field objects to xarray.Dataset

    Input: dictionary of climlab.Field objects
    (e.g. process.state or process.diagnostics dictionary)

    Output: xarray.Dataset object with all spatial axes,
    including 'bounds' axes indicating cell boundaries in each spatial dimension.

    Any items in the dictionary that are not instances of climlab.Field
    are ignored.'''
    from climlab.domain.field import Field

    ds = Dataset()
    for name, field in state.items():
        if isinstance(field, Field):
            ds[name] = Field_to_xarray(field)
            dom = field.domain
            for axname, ax in dom.axes.items():
                bounds_name = axname + '_bounds'
                ds.coords[bounds_name] = DataArray(ax.bounds, dims=[bounds_name],
                                    coords={bounds_name:ax.bounds})
                try:
                    ds[bounds_name].attrs['units'] = ax.units
                except:
                    pass
        else:
            warnings.warn('{} excluded from Dataset because it is not a Field variable.'.format(name))
    return ds

def to_xarray(input):
    '''Convert climlab input to xarray format.

    If input is a climlab.Field object, return xarray.DataArray

    If input is a dictionary (e.g. process.state or process.diagnostics),
    return xarray.Dataset object with all spatial axes,
    including 'bounds' axes indicating cell boundaries in each spatial dimension.

    Any items in the dictionary that are not instances of climlab.Field
    are ignored.'''
    from climlab.domain.field import Field
    if isinstance(input, Field):
        return Field_to_xarray(input)
    elif isinstance(input, dict):
        return state_to_xarray(input)
    else:
        raise TypeError('input must be Field object or dictionary of Field objects')
